use anyhow::{Error, Ok, Result};
use configparser::ini::Ini;
use lexopt::Parser;
use log::warn;

const GPCONF_PATH: &str = "/etc/gpservice/gp.conf";
const DEFAULT_SECTION: &str = "*";
const PROP_OPENCONNECT_ARGS: &str = "openconnect-args";

/// A struct representing the CLI arguments for the `openconnect` command.
/// Supports most of the options from the `openconnect` command line.
#[derive(Debug, Default)]
pub(crate) struct OpenconnectArgs {
    script: Option<String>,
    certificate: Option<String>,
    servercert: Option<String>,
}

impl OpenconnectArgs {
    pub fn script(&self) -> Option<String> {
        self.script.to_owned()
    }

    pub fn certificate(&self) -> Option<String> {
        self.certificate.to_owned()
    }

    pub fn servercert(&self) -> Option<String> {
        self.servercert.to_owned()
    }
}

/// Read the `gp.conf` file and return the `openconnect` arguments for the given server.
/// If the server is not found, the default section is used.
pub(crate) fn read_conf(server: &str) -> Result<OpenconnectArgs> {
    read_conf_from(GPCONF_PATH, server)
}

/// Private function to read the `openconnect` arguments for the given server from a given path.
/// Make it easy to write tests.
fn read_conf_from(path: &str, server: &str) -> Result<OpenconnectArgs> {
    let mut config = Ini::new();

    config.set_default_section(DEFAULT_SECTION);
    config.set_multiline(true);

    config.load(path).map_err(Error::msg)?;

    let default_openconnect_config = config
        .get(DEFAULT_SECTION, PROP_OPENCONNECT_ARGS)
        .unwrap_or_default();
    let server_openconnect_config = config
        .get(server, PROP_OPENCONNECT_ARGS)
        .unwrap_or(default_openconnect_config);

    let args = shlex::split(&server_openconnect_config).unwrap_or_default();
    parse_args(&args)
}

fn parse_args(args: &Vec<String>) -> Result<OpenconnectArgs> {
    use lexopt::prelude::*;

    let mut parser = Parser::from_args(args);

    let mut script: Option<String> = None;
    let mut certificate: Option<String> = None;
    let mut servercert: Option<String> = None;
    while let Some(arg) = parser.next()? {
        match arg {
            Long("script") | Short('s') => {
                script = Some(parser.value()?.parse()?);
            }
            Long("certificate") | Short('c') => {
                certificate = Some(parser.value()?.parse()?);
            }
            Long("servercert") => {
                servercert = Some(parser.value()?.parse()?);
            }
            _ => {
                warn!("Ignoring unknown argument: {}", arg.unexpected());
            }
        }
    }

    Ok(OpenconnectArgs {
        script,
        certificate,
        servercert,
    })
}

#[cfg(test)]
mod tests {
    use std::io::Write;
    use tempfile::NamedTempFile;

    use super::*;

    // Macro to create a temporary file with the given content.
    macro_rules! tempfile {
        ($content:expr) => {{
            let mut file = NamedTempFile::new().unwrap();
            write!(file, $content).unwrap();
            file
        }};
    }

    #[test]
    fn test_config_not_found() {
        let args = read_conf_from("non-existent-file", "server");
        assert!(args.is_err());
    }

    #[test]
    fn test_default_config() {
        let file = tempfile!(
            r#"
[*]
openconnect-args=--script=/script.sh
"#
        );
        let path = file.path().to_str().unwrap();
        let args = read_conf_from(path, "any server").unwrap();

        assert_eq!(args.script, Some("/script.sh".to_string()));
        assert_eq!(args.certificate, None);
        assert_eq!(args.servercert, None);
    }

    #[test]
    fn test_server_config() {
        let file = tempfile!(
            r#"
[*]
openconnect-args=--script=/script.sh

[server]
openconnect-args=--certificate=/cert.pem
"#
        );
        let path = file.path().to_str().unwrap();
        let args = read_conf_from(path, "server").unwrap();

        assert_eq!(args.script, None);
        assert_eq!(args.certificate, Some("/cert.pem".to_string()));
        assert_eq!(args.servercert, None);
    }

    #[test]
    fn test_ignore_unknown_args() {
        let file = tempfile!(
            r#"
[*]
openconnect-args=--script=/script.sh --unknown-arg -c /cert.pem
"#
        );
        let path = file.path().to_str().unwrap();
        let args = read_conf_from(path, "server").unwrap();

        assert_eq!(args.script, Some("/script.sh".to_string()));
        assert_eq!(args.certificate, Some("/cert.pem".to_string()));
        assert_eq!(args.servercert, None);
    }
}
